#! /usr/bin/env python
"""
metric.py
"""

from photutils import Background2D
from photutils import StdBackgroundRMS, MedianBackground
from astropy.io import fits
import numpy as np


def return_backgrounds(image):
    bkg = Background2D(image, (10, 10), filter_size=1,
                       bkg_estimator=MedianBackground(sigma_clip=None),
                       bkgrms_estimator=StdBackgroundRMS(sigma_clip=None))
    return bkg.background, bkg.background_rms


def return_fom(filename):
    image = fits.open(filename)[0].data

    median_bkg, rms_bkg = return_backgrounds(image)

    term0 = np.abs(median_bkg - np.median(image))
    term1 = rms_bkg
    fom = np.sum(term0 < term1) / len(image.flatten())
    return fom



#!/usr/bin/env python
"""utils.py"""
from numpy import random as np_random
from numpy import arange as np_arange
from numpy import array as np_array
from glob import glob
from astropy.io import fits
from os import path as os_path
from os import remove as os_remove


def flatten_images(images):
    """Flattens images for use in 1D analysis"""

    image_shape = images[0].shape
    images_flattened = []
    for i, image in enumerate(images):
        if i % 10 == 0:
            print('Flattening image %i/%i' % (i, len(images)))
        images_flattened.append(image.flatten())
    images_flattened = np_array(images_flattened)

    return images_flattened, image_shape


def create_fits(image_name,
                data,
                header=None):
    """Creates a fits image with an optional header

    Uses the astropy.io.fits pacakge to create a fits image.
    WARNING : THIS WILL OVERWRITE ANY FILE ALREADY NAMED 'image_name'.
    """
    # Creates the Header Data Unit
    hdu = fits.PrimaryHDU(data)

    # Adds the 'header' if None is not selected
    if header is not None:
        hdu.header = header

    # Remove the image if it currently exists
    if os_path.exists(image_name):
        os_remove(image_name)

    # Write the fits image to disk
    hdu.writeto(image_name)


def generate_random_ds9_list(n_random=6):
    """ Generates a random list of images to be viewed in ds9
    Viewed with: ds9 -zscale $(<ds9.list) """

    fname_arr = glob('ztf*sciimg.clean.fits')
    fname_arr.sort()

    n_images = min(n_random, len(fname_arr))

    idx_arr = np_random.choice(np_arange(len(fname_arr)), n_images,
                               replace=False)
    with open('ds9.list', 'w') as f:
        for idx in idx_arr:
            f.write('%s\n' % fname_arr[idx])
            f.write('%s\n' % fname_arr[idx].replace('.clean', ''))



#!/usr/bin/env python
"""fringe.py"""
from astropy.io import fits
from numpy import median as np_median
from numpy import abs as np_abs
from numpy import zeros_like as np_zeros_like
from numpy import dot as np_dot
from numpy import sqrt as np_sqrt
from numpy import newaxis as np_newaxis
from numpy import load as np_load
from glob import glob
from sys import exit as sys_exit
from os import path as os_path
from fringez.utils import create_fits
from fringez.utils import flatten_images


def generate_fringe_map(image):
    """ Create a fringe map from a science image."""
    median = np_median(image)
    median_absdev = np_median(np_abs(image - median))
    pixel_5sigma_plus = median + (median_absdev * 1.48 * 5)
    pixel_5sigma_minus = median - (median_absdev * 1.48 * 5)

    fringe_map = np_zeros_like(image)
    fringe_map[:] = image[:]
    cond1 = fringe_map >= pixel_5sigma_plus
    cond2 = fringe_map <= pixel_5sigma_minus
    cond = cond1 + cond2
    fringe_map[cond] = median

    fringe_map -= median
    fringe_map /= median_absdev

    return fringe_map, median_absdev


def gather_flat_fringe_maps():
    """Gathers all of the fringe images in the directory,
    flattened for 1D analysis"""
    fname_arr, fringes, rcid = gather_fringe_maps()
    fringe_maps_flattened, image_shape = flatten_images(fringes)
    return fname_arr, fringe_maps_flattened, image_shape, rcid


def gather_fringe_maps():
    """Gathers all of the science images in the directory and returns a list
    of the fringe maps, as well as the rcid of the science images."""

    fname_arr = glob('ztf*sciimg.fits')
    fname_arr.sort()

    with fits.open(fname_arr[0]) as f:
        header = f[0].header

    if 'RCID' in header:
        rcid = int(header['RCID'])
    else:
        rcid = 0

    image_shape = None
    fringe_maps = []
    for i, fname in enumerate(fname_arr):
        if i % 10 == 0:
            print('Generating fringe image %i/%i' % (i, len(fname_arr)))
        with fits.open(fname) as f:
            image = f[0].data
        fringe_map, _ = generate_fringe_map(image)

        if i == 0:
            image_shape = fringe_map.shape
        else:
            if fringe_map.shape != image_shape:
                print('%s != %s' % (str(fringe_map.shape), str(image_shape)))
                print('** ALL FRINGE MAPS MUST BE THE SAME SIZE **')
                print('** EXITING **')
                sys_exit(0)

        fringe_maps.append(fringe_map)

    return fname_arr, fringe_maps, rcid


def append_eigenvalues_to_header(header, fringe_ica):
    for i, eigenvalue in enumerate(fringe_ica.T):
        header['PCAEIG%02d' % i] = eigenvalue[0]

    return header


def calculate_fringe_bias(fringe_map, fringe_model):
    """ Generates fringe bias image for the provided science image.
    These formulas are taken from the scikit-learn estimator's
    transform and inverse transform methods.

    Models are loaded from disk as
    fringe_{MODEL_NAME}_comp{N_COMPONENTS}.c{CID}_q{QID}.{DATE}.model """

    fringe_map = fringe_map.flatten()
    fringe_map_transposed = fringe_map.reshape(1, len(fringe_map))

    mean = fringe_model['mean']
    components = fringe_model['components']
    explained_variance = fringe_model['explained_variance']

    fringe_map_transposed = fringe_map_transposed - mean
    fringe_ica = np_dot(fringe_map_transposed, components.T)
    fringe_ica /= np_sqrt(explained_variance)

    fringe_bias = np_dot(fringe_ica, np_sqrt(
        explained_variance[:, np_newaxis]) * components) + mean

    return fringe_bias, fringe_ica


def remove_fringe(image_name,
                  fringe_model_name,
                  debugFlag=False):
    """ Subtracts the fringe bias image from the science image, resulting in
    a clean image with extension *sciimg.clean.fits.

    Models are loaded from disk as
    fringe_{MODEL_NAME}_comp{N_COMPONENTS}.c{CID}_q{QID}.{DATE}.model """

    if not os_path.exists(image_name):
        print('Image missing! Exiting...')
        sys_exit(0)

    if not os_path.exists(fringe_model_name):
        print('Fringe model missing! Exiting...')
        sys_exit(0)

    print('Generating clean image for %s' % image_name)

    with fits.open(image_name) as f:
        image = f[0].data
        header = f[0].header
        image_shape = image.shape

    fringe_map, median_absdev = generate_fringe_map(image)

    fringe_model = np_load(fringe_model_name)

    fringe_bias, fringe_ica = calculate_fringe_bias(fringe_map, fringe_model)
    fringe_bias = fringe_bias.reshape(image_shape)
    fringe_bias *= median_absdev

    header = append_eigenvalues_to_header(header, fringe_ica)
    header['FRNGMDL'] = os_path.basename(fringe_model_name)

    image_clean = image - fringe_bias
    image_clean_fname = image_name.replace('.fits', '.clean.fits')
    create_fits(image_clean_fname, image_clean, header)
    print('-- %s saved to disk' % image_clean_fname)

    if debugFlag:
        extension = os_path.basename(fringe_model_name).replace('.model',
                                                                '.bias')
        fname = image_name.replace('.fits', '.%s.fits' % extension)
        create_fits(fname, fringe_bias, header)

        print('-- %s saved to disk' % fname)
